package gliparameterizer;

import java.util.Vector;

/**
 * <p>Title: Trees</p>
 * <p>Description: Keeps the tree records.</p>
 * <p>Copyright: Copyright (c) 2004 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */

class AllTrees {

  /**AllometryPackage objects for each defined species*/
  protected Vector mp_oAllometry = new Vector(0);
  /**Data for species not defined in mp_oAllometry (an "other" category)*/
  protected AllometryPackage m_oOther = null;
  /**Length of the plot in the X (E-W) direction, in meters*/
  protected int m_iPlotXLength;
  /**Length of the plot in the Y (N-S) direction, in meters*/
  protected int m_iPlotYLength;
  /**Size of plot cells*/
  protected double m_fCellLength = 8.0;
  /**Array with all the trees.  The array size is number of X cells by number
   * of Y cells.  Each Vector has all the Tree objects that belong in that cell,
   * unsorted.*/
  protected Vector[][] mp_oAllTrees = null;

  /**
   * Constructor.
   * @param iXLength int Length of the plot, in the X (E-W) direction, in m.
   * @param iYLength int Length of the plot, in the Y (N-S) direction, in m.
   * @throws Exception if the plot lengths are not greater than 0.
   */
  public AllTrees(int iXLength, int iYLength) throws Exception {
    if (iXLength <= 0 || iYLength <= 0) {
      throw new Exception("Plot lengths must be greater than 0.");
    }
    m_iPlotXLength = iXLength;
    m_iPlotYLength = iYLength;

    //Declare the trees array
    int iNumXCells = (int)java.lang.Math.ceil((double)iXLength / m_fCellLength),
        iNumYCells = (int)java.lang.Math.ceil((double)iYLength / m_fCellLength),
        iX, iY;

    mp_oAllTrees = new Vector[iNumXCells][iNumYCells];
    for (iX = 0; iX < mp_oAllTrees.length; iX++) {
      for (iY = 0; iY < mp_oAllTrees[iX].length; iY++) {
        mp_oAllTrees[iX][iY] = new Vector(0);
      }
    }
  }

  /**
   * Gets the canopy radius coefficient for a species.
   * @param sSpecies String Name of species.
   * @return double Canopy radius coefficient.
   * @throws Exception if the species name isn't recognized and m_oOther is
   * null.
   */
  public double GetCanopyRadiusCoefficient(String sSpecies) throws Exception {
    double fVal = 0;
    int i;
    boolean bFound = false;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_sSpecies.equals(sSpecies)) {
        fVal = oSpData.m_fRadiusCoefficient;
        bFound = true;
        break;
      }
    }

   if (!bFound) {
     if (m_oOther != null) {
       fVal = m_oOther.m_fRadiusCoefficient;
     } else {
       throw new Exception("Unrecognized species \""+sSpecies+"\".");
     }
   }
   return fVal;
  }

  /**
   * Gets the max height for a species.
   * @param sSpecies String Name of species.
   * @return double Max height, in m.
   * @throws Exception if the species name isn't recognized and m_oOther is
   * null.
   */
  public double GetMaxHeight(String sSpecies) throws Exception {
    double fVal = 0;
    int i;
    boolean bFound = false;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_sSpecies.equals(sSpecies)) {
        fVal = oSpData.m_fMaxHeight;
        bFound = true;
        break;
      }
    }

   if (!bFound) {
     if (m_oOther != null) {
       fVal = m_oOther.m_fMaxHeight;
     } else {
       throw new Exception("Unrecognized species \""+sSpecies+"\".");
     }
   }
   return fVal;
  }

  /**
   * Gets the height constant for a species.
   * @param sSpecies String Name of species.
   * @return double Canopy radius coefficient.
   * @throws Exception if the species name isn't recognized and m_oOther is
   * null.
   */
  public double GetHeightConstant(String sSpecies) throws Exception {
    double fVal = 0;
    int i;
    boolean bFound = false;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_sSpecies.equals(sSpecies)) {
        fVal = oSpData.m_fHeightConst;
        bFound = true;
        break;
      }
    }

   if (!bFound) {
     if (m_oOther != null) {
       fVal = m_oOther.m_fHeightConst;
     } else {
       throw new Exception("Unrecognized species \""+sSpecies+"\".");
     }
   }
   return fVal;
  }

  /**
   * Gets the canopy depth coefficient for a species.
   * @param sSpecies String Name of species.
   * @return double Canopy radius coefficient.
   * @throws Exception if the species name isn't recognized and m_oOther is
   * null.
   */
  public double GetCanopyDepthCoefficient(String sSpecies) throws Exception {
    double fVal = 0;
    int i;
    boolean bFound = false;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_sSpecies.equals(sSpecies)) {
        fVal = oSpData.m_fCanopyDepthCoeff;
        bFound = true;
        break;
      }
    }

   if (!bFound) {
     if (m_oOther != null) {
       fVal = m_oOther.m_fCanopyDepthCoeff;
     } else {
       throw new Exception("Unrecognized species \""+sSpecies+"\".");
     }
   }
   return fVal;
  }

  /**
   * Gets the canopy radius exponent for a species.
   * @param sSpecies String Name of species.
   * @return double Canopy radius exponent.
   * @throws Exception if the species name isn't recognized and m_oOther is
   * null.
   */
  public double GetCanopyRadiusExponent(String sSpecies) throws Exception {
    double fVal = 0;
    int i;
    boolean bFound = false;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_sSpecies.equals(sSpecies)) {
        fVal = oSpData.m_fRadiusExponent;
        bFound = true;
        break;
      }
    }

   if (!bFound) {
     if (m_oOther != null) {
       fVal = m_oOther.m_fRadiusExponent;
     } else {
       throw new Exception("Unrecognized species \""+sSpecies+"\".");
     }
   }
   return fVal;
  }

  /**
   * Adds a tree to the correct place in mp_oAllTrees.
   * @param oTree Tree to add.
   * @throws Exception if the tree doesn't fit into the plot.
   */
  public void AddTree(Tree oTree) throws Exception {
    int iXCell = (int) java.lang.Math.floor(oTree.m_fX / m_fCellLength),
        iYCell = (int) java.lang.Math.floor(oTree.m_fY / m_fCellLength);

    if (iXCell < 0 || iXCell >= mp_oAllTrees.length) {
      throw new Exception ("A tree's coordinates do not fit within the plot.  X = "
                           + oTree.m_fX+ ", Y = " + oTree.m_fY + ", Tag = "+
                           oTree.m_iTag+".");
    }
    if (iYCell < 0 || iYCell >= mp_oAllTrees[iXCell].length) {
      throw new Exception ("A tree's coordinates do not fit within the plot.  X = "
                           + oTree.m_fX+ ", Y = " + oTree.m_fY + ", Tag = "+
                           oTree.m_iTag+".");
    }

    mp_oAllTrees[iXCell][iYCell].add(oTree);
  }

  /**
   * Gets the maximum height across all tree species.
   * @return double Max tree height, in m.
   */
  public double GetMaxAllTreeHeight() {
    double fVal = 0;
    int i;
    for (i = 0; i < mp_oAllometry.size(); i++) {
      AllometryPackage oSpData = (AllometryPackage)mp_oAllometry.elementAt(i);
      if (oSpData.m_fMaxHeight > fVal) {
        fVal = oSpData.m_fMaxHeight;
      }
    }
    if (m_oOther != null) {
      if (m_oOther.m_fMaxHeight > fVal) {
        fVal = m_oOther.m_fMaxHeight;
      }
    }
    return fVal;
  }


  /**
   * Finds all candidate trees for intersection by the ray.  This casts a
   * wide net and lots of trees that do not in fact intersect will be returned.
   * @param oRay GLIRay Ray to get intersectors for.
   * @return Vector A Vector of Tree objects that might intersect.
   */
  public Vector FindTrees(GLIRay oRay) {
      Vector oReturnTrees = new Vector(0);
      Tree oTree;
      double fXDist,
             fMaxSearchRad; //search radius for shading trees
      int iHomeY, //grid cell coords of home cell
          iMaxX, iMaxY, //grid cell coords of max cell to search
          iMinX, iMinY, //grid cell coords of min cell to search
                iX, iY, i; //loop counters

      //Find the search radius
      fMaxSearchRad = (GetMaxAllTreeHeight() - oRay.m_fHeight) *
           (1.0 / (java.lang.Math.tan(oRay.m_fAltitude))) + //reciprocal of tangent of altitude
           20.0; //very large value for canopy radius

      //Get the home grid cell of the target point
      iHomeY = (int)java.lang.Math.floor(oRay.m_fY / m_fCellLength);

      //Figure out our min. and max. Y search cells - for each Y cell we loop
      //through, we'll find the min. and max. X for that cell
      iMaxY = (int)java.lang.Math.ceil((oRay.m_fY + fMaxSearchRad) / m_fCellLength);
      iMinY = (int)java.lang.Math.floor((oRay.m_fY - fMaxSearchRad) / m_fCellLength);

      //Move from min to max Y
      for (iY = iMinY; iY <= iMaxY; iY++) {

        if (iY >= 0 && iY < mp_oAllTrees[0].length) {

          //we know XY of center, and Y of point on edge of circle - solve for X
          //X2 = X1 - square root(distance^2 - (Y1 - Y2)^2)
          //If this is the home Y cell, then the X dist = the radius
          if (iY == iHomeY) {
            iMaxX = (int)java.lang.Math.ceil( (oRay.m_fX + fMaxSearchRad) /
                                        m_fCellLength);
            iMinX = (int)java.lang.Math.floor( (oRay.m_fX - fMaxSearchRad) /
                                         m_fCellLength);
          }
          else {
            //When we are south of the target Y, add one to grid cell number so
            //we'll be calculating the northern edge
            if ( (iY * m_fCellLength) < oRay.m_fY)
              fXDist = fMaxSearchRad * fMaxSearchRad - ( (oRay.m_fY -
                  ( (iY + 1) * m_fCellLength)) *
                  (oRay.m_fY - ( (iY + 1) * m_fCellLength)));
            else
              fXDist = fMaxSearchRad * fMaxSearchRad - ( (oRay.m_fY -
                  (iY * m_fCellLength)) * (oRay.m_fY - (iY * m_fCellLength)));
            if (fXDist > 0) {
              fXDist = java.lang.Math.sqrt(fXDist);
              //fXDist is positive and negative solution
              iMaxX = (int)java.lang.Math.ceil( (oRay.m_fX + fXDist) / m_fCellLength);
              iMinX = (int)java.lang.Math.floor( (oRay.m_fX - fXDist) / m_fCellLength);
            }
            else { //a solution < 0 means that there's no triangle - there are no
              //cells to the east or west of the current one
              iMaxX = (int)java.lang.Math.floor(oRay.m_fX / m_fCellLength);
              iMinX = iMaxX;
            }
          }

          for (iX = iMinX; iX <= iMaxX; iX++) {

            if (iX < mp_oAllTrees.length && iX >= 0) {

              //Phew!  Now get everything in that cell
              for (i = 0; i < mp_oAllTrees[iX][iY].size(); i++) {
                oTree = (Tree) mp_oAllTrees[iX][iY].elementAt(i);
                if (oTree.m_fHeight > oRay.m_fHeight) {
                  oReturnTrees.add(oTree);
                }
              }
            }
          } //end of for (iX = iMinXGridCell; iX <= iMaxXGridCell; iX++)
        }
      } //end of for (iY = iMinYGridCell; iY <= iMaxYGridCell; iY++)
      return oReturnTrees;
  }


}
/**Class packaging up the allometry data for a single species*/
class AllometryPackage {

    /**Species name for this data*/
    public String m_sSpecies = "";
    /**Canopy radius coefficient*/
    public double m_fRadiusCoefficient = 0;
    /**Canopy radius exponent*/
    public double m_fRadiusExponent = 1;
    /**Maximum height, in meters*/
    public double m_fMaxHeight = 0;
    /**Height constant*/
    public double m_fHeightConst = 0;
    /**Canopy depth coefficient*/
    public double m_fCanopyDepthCoeff = 0;

    public AllometryPackage() {;}

    public AllometryPackage(String sSpecies, double fRadiusCoefficient, double fRadiusExponent, double fMaxHeight, double fHeightConst, double fCanopyDepthCoeff) {
      m_sSpecies = sSpecies;
      m_fRadiusCoefficient = fRadiusCoefficient;
      m_fRadiusExponent = fRadiusExponent;
      m_fMaxHeight = fMaxHeight;
      m_fHeightConst = fHeightConst;
      m_fCanopyDepthCoeff = fCanopyDepthCoeff;
    }

  }

