package gliparameterizer;

import java.util.Vector;
import java.io.FileWriter;
import java.io.FileReader;

/**
 * <p>Title: Analyzer</p>
 * <p>Description: Runs analysis.</p>
 * <p>Copyright: Copyright (c) 2004 Charles D. Canham</p>
 * <p>Company: Institute of Ecosystem Studies</p>
 * @author Lora E. Murphy
 * @version 1.0
 */

class Analyzer {

  /**All the trees*/
  protected AllTrees m_oAllTrees;

  /**An array of GLIRay objects to analyze.*/
  protected Vector mp_oGLIRays = new Vector(0);

  public Analyzer() {
  }

  /**
   * Reads in the tree file and puts the trees into m_oAllTrees.
   * @param sFileName String Filename to read.
   * @throws Exception if there is a problem opening the file or the data
   * cannot be understood.
   */
  protected void ReadTreeFile(String sFileName) throws Exception {
    FileReader oIn = null;
    try {
      Vector p_oLine;
      String sSpecies, sVal;
      double fDbh, fX, fY;
      int iTag;
      byte iDamageCode;

      oIn = new FileReader(sFileName);

      //Skip the header line
      ModelFileFunctions.SkipLine(oIn);

      //The format is six columns: Tag, Species, DBH, Damage Code, X, and Y.
      p_oLine = ModelFileFunctions.ReadLine(oIn);
      while (p_oLine.size() > 0) {

        if (p_oLine.size() != 6) {
          throw new Exception(
              "There is the wrong number of data columns in the file \"" +
              sFileName + "\", with the line beginning \"" +
              String.valueOf(p_oLine.elementAt(0)) + "\".");
        }

        //Get Tag
        sVal = (String) p_oLine.elementAt(0);
        iTag = new Integer(sVal).intValue();

        //Get species
        sSpecies = (String) p_oLine.elementAt(1);
        if (sSpecies.length() == 0) {
          throw new Exception("There is a missing species name in the file \"" +
                              sFileName + "\".");
        }

        //Get dbh
        try {
          sVal = (String) p_oLine.elementAt(2);
          fDbh = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the DBH value \"" +
                              sVal + "\".");
        }

        //Get damage code
        try {
          sVal = (String) p_oLine.elementAt(3);
          iDamageCode = new Integer(sVal).byteValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName + "\" could not be read.  I could not successfully interpret the damage code value \"" +
                              sVal + "\".");
        }

        //Get X
        try {
          sVal = (String) p_oLine.elementAt(4);
          fX = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the X value \"" +
                              sVal + "\".");
        }

        //Get Y
        try {
          sVal = (String) p_oLine.elementAt(5);
          fY = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the Y value \"" +
                              sVal + "\".");
        }

        //Add a new tree
        m_oAllTrees.AddTree(new Tree(sSpecies, fX, fY, fDbh, iTag, iDamageCode,
                                     m_oAllTrees));

        p_oLine = ModelFileFunctions.ReadLine(oIn);
      }
    }
    catch (java.io.IOException e) {
      throw new Exception("There was a problem writing the file \"" + sFileName +
                          "\".  Message: " + e.getMessage());
    }
    finally {
      if (oIn != null) {
        oIn.close();
      }
    }
  }

  /**
   * Reads the allometry file and sets up m_oAllTrees.
   * @param sFileName String File to read.
   * @throws Exception if the file cannot be read or contains bad data.
   */
  protected void ReadAllometryFile(String sFileName) throws Exception {
    FileReader oIn = null;
    try {
      Vector p_oLine;
      String sSpecies, sVal;
      double fCanradCoeff, fCanradExp, fMaxHeight, fHeightConst, fDepthCoeff;
      int iXPlotLength, iYPlotLength;

      oIn = new FileReader(sFileName);

      //First two lines are X and Y plot lengths
      p_oLine = ModelFileFunctions.ReadLine(oIn);
      if (p_oLine.size() < 2) {
        throw new Exception("Unexpected file format for file \"" + sFileName +
                            "\".  Couldn't find X plot length.");
      }
      sVal = (String) p_oLine.elementAt(1);
      iXPlotLength = new Double(sVal).intValue();

      p_oLine = ModelFileFunctions.ReadLine(oIn);
      if (p_oLine.size() < 2) {
        throw new Exception("Unexpected file format for file \"" + sFileName +
                            "\".  Couldn't find Y plot length.");
      }
      sVal = (String) p_oLine.elementAt(1);
      iYPlotLength = new Double(sVal).intValue();

      m_oAllTrees = new AllTrees(iXPlotLength, iYPlotLength);

      //Skip header line
      ModelFileFunctions.SkipLine(oIn);

      //The format is five columns: species, canopy radius coefficient, max
      //tree height, height constant, and canopy depth coefficient.
      p_oLine = ModelFileFunctions.ReadLine(oIn);
      while (p_oLine.size() > 0) {

        if (p_oLine.size() != 6) {
          throw new Exception(
              "There is the wrong number of data columns in the file \"" +
              sFileName + "\", with the line beginning \"" +
              String.valueOf(p_oLine.elementAt(0)) + "\".");
        }

        //Get species
        sSpecies = (String) p_oLine.elementAt(0);

        //Get canopy radius coefficient
        sVal = (String) p_oLine.elementAt(1);
        fCanradCoeff = new Double(sVal).doubleValue();

        //Get canopy radius exponent
        sVal = (String) p_oLine.elementAt(2);
        fCanradExp = new Double(sVal).doubleValue();

        //Get max tree height
        sVal = (String) p_oLine.elementAt(3);
        fMaxHeight = new Double(sVal).doubleValue();

        //Get height constant
        sVal = (String) p_oLine.elementAt(4);
        fHeightConst = new Double(sVal).doubleValue();

        //Get canopy depth coefficient
        sVal = (String) p_oLine.elementAt(5);
        fDepthCoeff = new Double(sVal).doubleValue();

        //Add a new allometry record
        //Is the species "other"?
        if (sSpecies.equalsIgnoreCase("other")) {
          m_oAllTrees.m_oOther = new AllometryPackage(sSpecies, fCanradCoeff,
              fCanradExp, fMaxHeight, fHeightConst, fDepthCoeff);
        }
        else {
          m_oAllTrees.mp_oAllometry.add(new AllometryPackage(sSpecies,
              fCanradCoeff, fCanradExp, fMaxHeight, fHeightConst, fDepthCoeff));
        }

        p_oLine = ModelFileFunctions.ReadLine(oIn);
      }
    }
    catch (java.io.IOException e) {
      throw new Exception("There was a problem writing the file \"" + sFileName +
                          "\".  Message: " + e.getMessage());
    }
    catch (java.lang.NumberFormatException e) {
      throw new Exception("Data in the file \"" + sFileName + "\" could not be read.  I could not successfully interpret an expected number value as a number.");
    }
    finally {
      if (oIn != null) {
        oIn.close();
      }
    }
  }

  /**
   * Reads the GLI ray file and populates mp_oGLIRays.
   * @param sFileName String File to read.
   * @throws Exception if the file cannot be read or contains invalid data.
   */
  protected void ReadGLIRayFile(String sFileName) throws Exception {
    FileReader oIn = null;
    try {
      Vector p_oLine;
      String sSubquad, sLight, sVal = null;
      double fAltitude, fAzimuth, fX, fY, fHeight;

      oIn = new FileReader(sFileName);

      //Skip the header line
      ModelFileFunctions.SkipLine(oIn);

      //The format is seven columns: Altitude, Azimuth, Subquad, Light, X, Y,
      //and height.
      p_oLine = ModelFileFunctions.ReadLine(oIn);
      while (p_oLine.size() > 0) {

        if (p_oLine.size() != 7) {
          throw new Exception(
              "There is the wrong number of data columns in the file \"" +
              sFileName + "\", with the line beginning \"" +
              String.valueOf(p_oLine.elementAt(0)) + "\".");
        }

        //Get altitude
        try {
          sVal = (String) p_oLine.elementAt(0);
          fAltitude = new Float(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the altitude value " +
                              sVal + ".");
        }

        //Get azimuth
        try {
          sVal = (String) p_oLine.elementAt(1);
          fAzimuth = new Float(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the azimuth value " +
                              sVal + ".");
        }

        //Get subquad
        sSubquad = (String) p_oLine.elementAt(2);

        //Get light level
        sLight = (String) p_oLine.elementAt(3);

        //Get X
        try {
          sVal = (String) p_oLine.elementAt(4);
          fX = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the X value " +
                              sVal + ".");
        }

        //Get Y
        try {
          sVal = (String) p_oLine.elementAt(5);
          fY = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the Y value " +
                              sVal + ".");
        }

        //Get height
        try {
          sVal = (String) p_oLine.elementAt(6);
          fHeight = new Double(sVal).doubleValue();
        }
        catch (java.lang.NumberFormatException e) {
          throw new Exception("Data in the file \"" + sFileName +
              "\" could not be read.  I could not successfully interpret the height value " +
                              sVal + ".");
        }

        //Add a new gli ray
        mp_oGLIRays.add(new GLIRay(fX, fY, fHeight, fAzimuth, fAltitude,
                                   sSubquad, sLight));

        p_oLine = ModelFileFunctions.ReadLine(oIn);
      }
    }
    catch (java.io.IOException e) {
      throw new Exception("There was a problem writing the file \"" + sFileName +
                          "\".  Message: " + e.getMessage());
    }
    finally {
      if (oIn != null) {
        oIn.close();
      }
    }
  }

  /**
   * Runs the analysis and prints out the results.
   * @param sAllometryFileName File name of the allometry in file.
   * @param sGLIRayFileName File name of the GLI ray in file.
   * @param sTreeFileName File name of the tree in file.
   * @param sOutFileName File name to which to write the output.
   * @throws Exception if the file couldn't be written.
   */
  protected void RunAnalysis(String sAllometryFileName, String sGLIRayFileName,
                             String sTreeFileName, String sOutFileName) throws
      Exception {
    FileWriter oOut = null;
    try {
      Vector p_oIntersectors;
      Tree oTree;
      GLIRay oRay;
      int i, j;

      ReadAllometryFile(sAllometryFileName);
      ReadGLIRayFile(sGLIRayFileName);
      ReadTreeFile(sTreeFileName);

      //Open the results file for writing
      oOut = new FileWriter(sOutFileName);
      oOut.write("ALT\tAZIMUTH\tSUBQUAD\tLIGHT\tRAY X\tRAY Y\tRAY HEIGHT\t# INTERCEPTORS\tTAG\tSPP\tDBH\tDAMCODE\n");

      //Analyze each ray
      for (i = 0; i < mp_oGLIRays.size(); i++) {
        oRay = (GLIRay) mp_oGLIRays.elementAt(i);

        //Get the trees that intersect this ray
        p_oIntersectors = GLICalcs.FindIntersectors(oRay, m_oAllTrees);

        //Write the ray and its intersectors
        oOut.write(String.valueOf(oRay.m_fAltitude * 180 / java.lang.Math.PI) +
                   "\t" +
                   String.valueOf(oRay.m_fAzimuth * 180 / java.lang.Math.PI) +
                   "\t" +
                   String.valueOf(oRay.m_sSubquad) + "\t" +
                   String.valueOf(oRay.m_sLight) + "\t" +
                   String.valueOf(oRay.m_fX) + "\t" +
                   String.valueOf(oRay.m_fY) + "\t" +
                   String.valueOf(oRay.m_fHeight) + "\t" +
                   String.valueOf(p_oIntersectors.size()));

        for (j = 0; j < p_oIntersectors.size(); j++) {
          oTree = (Tree) p_oIntersectors.elementAt(j);

          oOut.write("\t" + String.valueOf(oTree.m_iTag) +
                     "\t" + oTree.m_sSpecies +
                     "\t" + String.valueOf(oTree.m_fDBH) +
                     "\t" + String.valueOf(oTree.m_iDamageCode));
        }

        oOut.write("\n");
      }

    }
    catch (java.io.IOException e) {
      throw new Exception("There was a problem writing the file \"" +
                          sOutFileName + "\".  Message: " + e.getMessage());
    }
    finally {
      if (oOut != null) {
        oOut.close();
      }
    }
  }

}
